/*
 *	Buffer management
 */
 
#include <linuxmt/types.h>
#include <linuxmt/sched.h>
#include <linuxmt/fs.h>

static struct buffer_head buffer[NR_BUFFERS];
static struct wait_queue *buf_free_q=NULL;

/*
 *	STUB
 */
void blkdev_queue(a,b)
struct buffer_head *a;
int b;
{
	;
}
/*
 **************************
 */

void buffer_wait(block)
struct buffer_head *block;
{
	while(!(block->b_state&BH_COMPLETE))
		sleep_on(&block->b_waitq);
}

void buffer_wait_clean(block)
struct buffer_head *block;
{
	while(block->b_state&BH_DIRTY)
		sleep_on(&block->b_waitq);
}

/*
 *	Get a block from the buffer cache. Firstly try and
 *	find it in the cache, secondly try to use an empty
 *	slot, thirdly evict an existing item that is clean
 *	fourthly evict an item and write it back.
 *
 *	The returned buffer has been locked, and may have 
 *	an error if the I/O failed. 
 *
 *	Set flags to BG_NOREAD if the buffer is full sized
 *	and you are writing (ie you don't need the old
 *	contents first).
 *
 *	pri is 1 for a normal I/O . 0 for a swap. This is so
 *	10 processes swapped out and reading don't deadlock on
 *	the copy in/out from swap disk.
 */
 
struct buffer_head *buffer_get(dev,block,flags, pri)
kdev_t dev;
unsigned long block;
int flags;
int pri;
{
	struct buffer_head *bp;
	struct buffer_head *freep;
	struct buffer_head *freeable;
	int i;
retry:
	freep=NULL;
	freeable=NULL;
	bp=&buffer[0];
	for(i=pri;i<NR_BUFFERS;i++)
	{
		/*
		 *	If it has data in it and its not being reclaimed
		 *	it may be what we want. If its being reclaimed we
		 *	are too late and just missed the cache entry.
		 */
		if((bp->b_state&BH_USED) && !(bp->b_state&BH_RECLAIM))	
		{
			if(bp->b_blocknr==block && bp->b_dev==dev)
			{
				bp->b_locks++;
				/*
				 *	Wait if not yet filled.
				 */
				buffer_wait(bp);
				return bp;
			}
			/*
			 *	Freeable buffer if no free and not in cache
			 */
			if(bp->b_locks==0)
			{
				if(!freeable || !(bp->b_state&BH_DIRTY))
					freeable=bp;
			}
		}
		else
			freep=bp;
	}
	/*
	 *	Not in cache
	 */
	 
	if(freep==NULL)
	{
		if(freeable)
		{
			/*
			 *	This may well sleep. Thus we want to 
			 *	mark the buffer BH_RECLAIM so nobody can
			 *	allocate it while we flush it.
			 */
			freeable->b_locks++;
			freeable->b_state|=BH_RECLAIM;
			blkdev_queue(freeable, BLOCK_WRITE);
			buffer_wait_clean(freeable);
			freeable->b_state&=~BH_RECLAIM;
			freep=freeable;
		}
		else
		{
			/*
			 *	Wait for a free buffer	
			 */
			sleep_on(&buf_free_q);
			goto retry;
		}
	}
	else
		freep->b_locks++;

	freep->b_dev=dev;
	freep->b_blocknr=block;
	freep->b_state=BH_USED;
	freep->b_error=0;

	/*
	 *	Do we want the previous data ?
	 */
	 	
	if(!(flags&BG_NOREAD))
	{
		blkdev_queue(freep, BLOCK_READ);
		/*
		 *	Wait for a result
		 */
		buffer_wait(freep);
		if(freep->b_error)
			return freep;
	}
	
	/*
	 *	Either read or don't care.
	 */
	 
	return freep;
}

void buffer_free(bp)
struct buffer_head *bp;
{
	bp->b_locks--;
	if(bp->b_locks==-1)
		panic("buffer_free:over");
	/*
	 *	Wake any buffer starved process.
	 */
	if(!bp->b_locks)
		wake_up(&buf_free_q);
}

void buffer_lock(bp)
struct buffer_head *bp;
{
	bp->b_locks++;
}

void buffer_unlock(bp)
struct buffer_head *bp;
{
	bp->b_locks--;
	if(bp->b_locks==-1)
		panic("buffer_unlock: over");
}

/*
 *	The sync system call. Write all the blocks to disk. We will need
 *	to write all the inodes back first once we are done.
 */
 	
int sys_sync()
{
	struct buffer_head *bp=&buffer[0];
	int i;
/*	inode_sync();	*/	/* Sync inodes to blocks */
	/*
	 *	The remaining blocks to disk.
	 */
	for(i=0;i<10;i++)
	{
		if(bp->b_state&(BH_USED|BH_DIRTY)==(BH_USED|BH_DIRTY))
		{
			if(!(bp->b_state&BH_RECLAIM))
				blkdev_queue(bp, BLOCK_WRITE);
		}
		bp++;
	}
	return 0;
}

/*
 *	Set up the buffer cache 
 */
 
void buffer_init()
{
	int i;
	for(i=0;i< NR_BUFFERS;i++)
	{
		buffer[i].b_state=0;
	}
	printk("Buffer cache: %d buffers.\n",NR_BUFFERS);
}
